/* 
 *  XMMS Crossfade Plugin
 *  Copyright (C) 2000  Peter Eisenlohr <p.eisenlohr@gmx.net>
 *
 *  based on the original OSS Output Plugin
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#include "config.h"

#include "configure.h"
#include "convert.h"
#include "crossfade.h"
#include "format.h"
#include "monitor.h"
#include "oss.h"
#include "rate.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <ctype.h>

#include "interface.h"
#include "support.h"


/* output plugin callback prototypes */
static void xfade_init          ();
static void xfade_set_volume    (int  l, int  r);
static void xfade_get_volume    (int *l, int *r);
static gint xfade_open_audio    (AFormat fmt, int rate, int nch);
static void xfade_write_audio   (void *ptr, int length);
static void xfade_close_audio   ();
static void xfade_flush         (int time);
static void xfade_pause         (short paused);
static gint xfade_buffer_free   ();
static gint xfade_buffer_playing();
static gint xfade_written_time  ();
static gint xfade_output_time   ();

/* output plugin callback table */
static OutputPlugin xfade_op =
{
  NULL,
  NULL,
  "Crossfade Plugin " VERSION,
  xfade_init,
  xfade_about,
  xfade_configure,
  xfade_get_volume,
  xfade_set_volume,
  xfade_open_audio,
  xfade_write_audio,
  xfade_close_audio,
  xfade_flush,
  xfade_pause,
  xfade_buffer_free,
  xfade_buffer_playing,
  xfade_output_time,
  xfade_written_time,
};

/* internal prototypes */
static void *buffer_thread_f(void *arg);
static void  sync_output    ();

#ifdef HAVE_CTOR
void fini_dtor() __attribute__((destructor));
#else
void _fini();
#endif


/* local variables */
static gboolean  realtime;
static gboolean  is_http;

static gint64    streampos;  /* position within current song (input bytes) */
static gboolean  playing;
       gboolean  opened;     /* TRUE after cf_open(), FALSE after cf_close() */
static gboolean  paused;     /* TRUE: no playback (but still filling buffer) */
static gboolean  stopped;    /* TRUE: stop buffer thread ASAP */

#ifdef HAVE_OSS   /* avoid 'defined but not used' */
static plugin_config_t  default_op_config = DEFAULT_OP_CONFIG;
#endif
static plugin_config_t  the_op_config;
       OutputPlugin    *the_op = NULL;
       gboolean output_opened = FALSE;
static gint     output_offset = 0;

/*
 *  Available fade configs:
 * 
 *  fc_start:       First song, only in_len and in_level are used
 *  fc_xfade:       Automatic crossfade at end of song
 *  fc_album:       Like xfade but for consecutive songs of the same album
 *  fc_manual:      Manual crossfade (triggered by Next or Prev)
 *  fc_stop(?):     Last song, only out_len and out_level are used
 *                  (hard to implement, will probably be removed)
 *
 *  NOTE: As of version 0.2 of xmms-crossfade,
 *        only xfade and manual are implemented.
 *         
 *
 *  Available parameters:
 * 
 *              | start | xfade | manual | stop
 *  ------------+-------+-------+--------+------
 *  in_len      |   yes |   yes |    yes |   no
 *  in_volume   |   yes |   yes |    yes |   no
 *  offset      |    no |   yes |    yes |   no
 *  out_len     |    no |   yes |    yes |  yes
 *  out_volume  |    no |   yes |    yes |  yes
 *  flush       |  * no |  * no |  * yes | * no
 *  ------------+-------+-------+--------+------
 *  out_trigger |    no |   yes |     no |  yes
 *
 *  Items marked with (*) are not configureable by the user.
 *
 */

static struct timeval last_close;
static struct timeval last_write;

static format_t in_format;
static format_t out_format;

static buffer_t  the_buffer;
       buffer_t *buffer = &the_buffer;

       pthread_mutex_t buffer_mutex;
static pthread_t       buffer_thread;

static convert_context_t  convert_context;
static rate_context_t     rate_context;

static config_t  the_config;
       config_t *config = &the_config;
       config_t  config_default = CONFIG_DEFAULT;

static fade_config_t *fade_config = NULL;


/* this is the entry point for XMMS */
OutputPlugin *get_oplugin_info()
{
  return &xfade_op;
}

static gint
output_list_f(gconstpointer a, gconstpointer b)
{
  OutputPlugin *op = (OutputPlugin *)a;
  gchar      *name = (gchar *)b;

  return strcmp(g_basename(op->filename), name);
}

static OutputPlugin *
find_output()
{
  GList        *list, *element;
  OutputPlugin *op = NULL;

  /* find output plugin */
  if(config->output_method == OUTPUT_METHOD_BUILTIN_OSS) {
#ifdef HAVE_OSS
    op = xfade_get_builtin_oss_oplugin_info();
    the_op_config = default_op_config;
#else
    DEBUG(("[crossfade] find_output: no OSS support compiled in!\n"));
#endif
  }
  else if(config->output_method == OUTPUT_METHOD_PLUGIN) {
    if(config->op_name && (list = get_output_list()))
      if((element = g_list_find_custom(list, config->op_name, output_list_f)))
	op = element->data;
    
    if(op == &xfade_op) {
      DEBUG(("[crossfade] find_output: can't use myself as output plugin!\n"));
      op = NULL;
    }
    else if(!op) {
      DEBUG(("[crossfade] find_output: could not find output plugin \"%s\"\n",
	     config->op_name ? config->op_name : "#NULL#"));
    }
    else  /* ok, we have a plugin. last, get its compatibility options */
      xfade_load_plugin_config(config->op_config_string, config->op_name,
			       &the_op_config);
  }
  else if(config->output_method == OUTPUT_METHOD_BUILTIN_NULL) {
    DEBUG(("[crossfade] find_output: Null Output not yet implemented!\n"));
  }
  else {
    DEBUG(("[crossfade] find_output: unknown output method %d!\n",
	   config->output_method));
  }

  return op;
}

static gint
open_output()
{
  pthread_attr_t attr;

  /* sanity check */
  if(output_opened)
    DEBUG(("[crossfade] open_output: WARNING: output_opened=TRUE!\n"));

  /* reset */
  output_opened = FALSE;
  output_offset = 0;

  /* get output plugin (this will also init the_op_config) */
  if(!(the_op = find_output())) {
    DEBUG(("[crossfade] open_output: could not find any output!\n"));
    return -1;
  }

  /* print output plugin info */
  DEBUG(("[crossfade] open_output: using \"%s\" for output",
	 the_op->description ? the_op->description : "#NULL#"));

  if(realtime)
    DEBUG((" (RT)"));

  if(the_op_config.throttle_enable) 
    DEBUG((realtime
	   ? " (throttled (disabled with RT))"
	   : " (throttled)"));

  if(the_op_config.max_write_enable)
    DEBUG((" (max_write=%d)", the_op_config.max_write_len));

  DEBUG(("\n"));

  /* open plugin */
  if(!the_op->open_audio(out_format.fmt, out_format.rate, out_format.nch)) {
    DEBUG(("[crossfade] open_output: open_audio() failed!\n"));
    the_op = NULL;
    return -1;
  }

  /* clear buffer struct */
  memset(buffer, 0, sizeof(* buffer));
  
  /* calculate buffer size */
  buffer->mix_size     = MS2B(xfade_mix_size_ms(config)) & -4;
  buffer->sync_size    = MS2B(config->sync_size_ms) & -4;
  buffer->preload_size = MS2B(config->preload_size_ms) & -4;
  
  buffer->size = (buffer->mix_size +     /* mixing area */
		  buffer->sync_size +    /* additional sync */
		  buffer->preload_size); /* preload */
  
  DEBUG(("[crossfade] open_output: buffer: size=%d (%d+%d+%d=%d ms)\n",
	 buffer->size,
	 B2MS(buffer->mix_size),
	 B2MS(buffer->preload_size),
	 B2MS(buffer->sync_size),
	 B2MS(buffer->size)));
  
  /* allocate buffer */
  if(!(buffer->data = g_malloc0(buffer->size))) {
    DEBUG(("[crossfade] open_output: error allocating buffer!"));
    the_op->close_audio();
    the_op = NULL;
    return -1;
  }
  
  /* rewind buffer */
  buffer->rd_index = 0;
  buffer->used     = 0;
  buffer->preload  = buffer->preload_size;

  /* make sure stopped is TRUE -- otherwise the buffer thread would
   * stop again immediatelly after it has been started. */
  stopped = FALSE;
  
  /* create and run buffer thread */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  if(pthread_create(&buffer_thread, &attr, buffer_thread_f, NULL)) {
    PERROR("[crossfade] open_output: pthread_create()");
    pthread_attr_destroy(&attr);
    g_free(buffer->data);
    the_op->close_audio();
    the_op = NULL;
    return -1;
  }
  pthread_attr_destroy(&attr);

  /* start updating monitor */
  xfade_start_monitor();

  /* done */
  output_opened = TRUE;
  return 0;
}

static void
xfade_init()
{
  /* load config */
  memset(config, 0, sizeof(*config));
  *config = config_default;
  xfade_load_config();

  /* set default strings if there is no existing config */
  if(!config->oss_alt_audio_device)
    config->oss_alt_audio_device = g_strdup(DEFAULT_OSS_ALT_AUDIO_DEVICE);
  if(!config->oss_alt_mixer_device)
    config->oss_alt_mixer_device = g_strdup(DEFAULT_OSS_ALT_MIXER_DEVICE);
  if(!config->op_config_string)
    config->op_config_string = g_strdup(DEFAULT_OP_CONFIG_STRING);
  if(!config->op_name)
    config->op_name = g_strdup(DEFAULT_OP_NAME);

  /* check for realtime priority, it needs some special attention */
  realtime = xmms_check_realtime_priority();

  /* show monitor win if enabled in config */
  xfade_check_monitor_win();

  /* create buffer mutex (never fails) */
  pthread_mutex_init(&buffer_mutex, NULL);

  /* setup out_format */
  setup_format(FMT_S16_NE, OUTPUT_RATE, OUTPUT_NCH, &out_format);

  /* init contexts */
  convert_init(&convert_context);
  rate_init(&rate_context);

  /* reset */
  stopped = FALSE;

  /* find current output plugin early so that volume control works
   * even if playback has not started yet. */
  if(!(the_op = find_output()))
    DEBUG(("[crossfade] init: could not find any output!\n"));
}

#ifdef HAVE_CTOR
void fini_dtor()
#else
void _fini()
#endif
{
  DEBUG(("[crossfade]\n"));
  DEBUG(("[crossfade] _fini: cleanup:\n"));

  /* signal buffer_thread to stop */
  pthread_mutex_lock(&buffer_mutex);
  stopped = TRUE;
  paused  = FALSE;
  
  /* wait for buffer thread to clean up and terminate */
  while(output_opened) {
    pthread_mutex_unlock(&buffer_mutex);
    xmms_usleep(10000);
    pthread_mutex_lock(&buffer_mutex);
  }
  xmms_usleep(10000);

  /* unlock and free buffer mutex */
  pthread_mutex_unlock(&buffer_mutex);
  pthread_mutex_destroy(&buffer_mutex);

  /* free resources */
  rate_free(&rate_context);
  convert_free(&convert_context);
  if(config->oss_alt_audio_device) g_free(config->oss_alt_audio_device);
  if(config->op_name) g_free(config->op_name);
  xfade_free_config();

  DEBUG(("[crossfade] _fini: done.\n"));
}

void
xfade_get_volume(int *l, int *r)
{
  if(the_op && the_op->get_volume) {
    if(config->mixer_reverse)
      the_op->get_volume(r, l);
    else
      the_op->get_volume(l, r);
  }
}

void
xfade_set_volume(int l, int r)
{
  if(!config->enable_mixer) return;
  if(the_op && the_op->set_volume) {
    if(config->mixer_reverse)
      the_op->set_volume(r, l);
    else
      the_op->set_volume(l, r);
  }
}

static gint
xfade_open_audio(AFormat fmt, int rate, int nch)
{
  gint   session_id;
  gint   pos;
  gchar *title;
  gchar *file;

  struct timeval tv;
  glong          dt;

  DEBUG(("[crossfade]\n"));

  session_id = ctrlsocket_get_session_id();
  pos   = xmms_remote_get_playlist_pos  (session_id);
  title = xmms_remote_get_playlist_title(session_id, pos);
  file  = xmms_remote_get_playlist_file (session_id, pos);

#if 0
  DEBUG(("[crossfade] open_audio: session=%d pos=%d\n", session_id, pos));
  DEBUG(("[crossfade] open_audio: title=\"%s\"\n", title));
  DEBUG(("[crossfade] open_audio: file=\"%s\"\n", file));
#endif
  
  /* check for HTTP streaming */
  is_http = config->enable_http_workaround &&
    (0 == strncasecmp(file, "http://", 7));
  
  if(is_http)
    DEBUG(("[crossfade] open_audio: HTTP underrun workaround enabled.\n"));
  
  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  /* reset writer timeout */
  gettimeofday(&last_write, NULL);

  /* calculate time since last close() (don't care about overflows at 24h) */
  if(output_opened) {
    gettimeofday(&tv, NULL);
    dt = (tv.tv_sec - last_close.tv_sec)  * 1000
      + (tv.tv_usec - last_close.tv_usec) / 1000;
  }
  else dt = 0;

  DEBUG(("[crossfade] open_audio: fmt=%s rate=%d nch=%d dt=%ld ms\n",
	 format_name(fmt), rate, nch, dt));
  
  /* check format */
  if(setup_format(fmt, rate, nch, &in_format) < 0) {
    DEBUG(("[crossfade] open_audio: format not supported!\n"));
    return 0;
  }

  /* (re)open the device if necessary */
  if(!output_opened) {
    if(open_output()) {
      DEBUG(("[crossfade] open_audio: error opening/configuring output!\n"));
      pthread_mutex_unlock(&buffer_mutex);
      return 0;
    }
    fade_config = NULL;
  }

  /* reset */
  streampos = 0;
  playing   = TRUE;
  opened    = TRUE;
  paused    = FALSE;

  /* reset mixer; initialize leading gap killer */
  buffer->mix       = 0;
  buffer->fade      = 0;
  buffer->gap_level = config->gap_lead_level;
  buffer->gap_len   = (config->gap_lead_enable
		       ? MS2B(config->gap_lead_len_ms) & -4 : 0);
  buffer->gap       = buffer->gap_len;
  buffer->written   = 0;
  buffer->reopen    = -1;
  
  /*
   * Example 1: offset < 0 --> mix streams together
   * Example 2: offset > 0 --> insert pause between streams
   * 
   *     |----- out_len -----|    *     |out_len|
   *     |                   |    *     |       |
   * ~~~~~-_        /T~~~~~~~T~~  * ~~~~~\      |               /T~~
   *        ~-_    / |       |    *       \     |              / |
   *           ~-_/  |       |    *        \    |             /  |
   *             /~-_|       |    *         \   |            /   |
   *            /    T-_     |    *          \  |           /    |
   *           /     |  ~-_  |    *           \ |          /     |
   * _________/______|_____~-|__  * ___________\__________/______|__
   *          |in_len|       |    *             |         |in_len|
   *          |<-- offset ---|    *             |offset-->|
   *
   *  a) avail:   max(0, used - preload)
   *  b) out_len: 0 .. avail
   *  c) in_len:  0 .. #
   *  d) offset:  -avail .. buffer->mix_size - out_size
   *  e) skip:    min(used, preload)
   *
   */

  /* start mixing */
  if(!fade_config) {
    /* sanity */
  }
  else if(fade_config->type == FADE_TYPE_FLUSH) {
    DEBUG(("[crossfade] open_audio: FLUSH:\n"));

    /* down the toilet */
    the_op->flush(0);

    /* rewind buffer */
    buffer->rd_index = 0;
    buffer->used     = 0;
    buffer->preload  = buffer->preload_size;  
  }
  else if(fade_config->type == FADE_TYPE_REOPEN) {
    DEBUG(("[crossfade] open_audio: REOPEN:\n"));

    if(fade_config->flush) {
      /* close/reopen device immediately */
      if(the_op->close_audio) the_op->close_audio();
      if(!the_op->open_audio(out_format.fmt, out_format.rate, out_format.nch)) {
	DEBUG(("[crossfade] open_audio: REOPEN: reopening output plugin failed!\n"));
	g_free(buffer->data);
	output_opened = FALSE;
	pthread_mutex_unlock(&buffer_mutex);
	return 0;
      }
      
      /* rewind buffer */
      buffer->rd_index = 0;
      buffer->used     = 0;
      buffer->preload  = buffer->preload_size;  
    }
    else
      buffer->reopen = buffer->used;
  }
  else if(xfade_cfg_use_xfade(fade_config)) {
    gint avail, out_len, in_len, offset, skip;
    gint index, length, fade, n;
    gfloat out_scale, in_scale;
    gboolean out_len_clipped = FALSE, offset_clipped = FALSE;

    out_scale = 1.0f - (gfloat)xfade_cfg_fadeout_volume(fade_config) / 100.0f;
    in_scale  = 1.0f - (gfloat)xfade_cfg_fadein_volume (fade_config) / 100.0f;

    /* rules (see above) */
    /* a */
    avail = buffer->used - buffer->preload_size;
    if(avail < 0) avail = 0;

    /* b */
    out_len = MS2B(xfade_cfg_fadeout_len(fade_config)) & -4;
    if(out_len < 0) out_len = 0;
    if(out_len > avail) {
      out_len = avail;
      out_len_clipped = TRUE;
    }

    /* c */
    in_len = MS2B(xfade_cfg_fadein_len(fade_config)) & -4;
    if(in_len < 0) in_len = 0;

    /* d */
    offset = MS2B(xfade_cfg_offset(fade_config)) & -4;
    if(offset < -avail) {
      offset = -avail;
      offset_clipped = TRUE;
    }
    if(offset > (buffer->mix_size - out_len))
      offset = buffer->mix_size - out_len;

    /* e */
    skip = buffer->preload_size;
    if(skip > buffer->used) skip = buffer->used;

    /* cut off rest of stream (decreases latency on manual songchange) */
    if(fade_config->flush) {
      gint cutoff = avail - MAX(out_len, -offset);  /* MAX() -> glib.h */
      if(cutoff > 0) {
	DEBUG(("[crossfade] open_audio: XFADE: %d ms flushed\n", B2MS(cutoff)));
	buffer->used -= cutoff;
	avail        -= cutoff;
      }
    }

    /* begin modifying buffer at index */
    index = (buffer->rd_index + buffer->used - out_len) % buffer->size;

    /* fade out (modifies buffer directly) */
    fade = 0;
    length = out_len;
    while(length > 0) {
      gint16 *p = buffer->data + index;
      gint blen = buffer->size - index;
      if(blen > length) blen = length;

      for(n = blen/4; n > 0; n--) {
	gfloat factor = 1.0f - (((gfloat)fade / out_len) * out_scale);
	*p++ = (gfloat)*p * factor;
	*p++ = (gfloat)*p * factor;
	fade += 4;
      }

      index = (index + blen) % buffer->size;
      length -= blen;
    }

    /* add silence (modifies buffer directly) */
    if(offset > 0) {
      length = offset;
      while(length > 0) {
	gint16 *p = buffer->data + index;
	gint blen = buffer->size - index;
	if(blen > length) blen = length;

	memset(p, 0, blen);

	index = (index + blen) % buffer->size;
	length -= blen;
      }
      buffer->used += offset;
    }

    /* start fading */
    if(in_len > 0) {
      buffer->fade = in_len;
      buffer->fade_len = in_len;
      buffer->fade_scale = in_scale;
    }
    else buffer->fade = 0;

    /* start mixing */
    if(offset < 0) {
      length = -offset;
      buffer->mix = length;
      buffer->used -= length;
    }
    else buffer->mix = 0;

    /* set reopen countdown. after buffer_thread_f has written
     * buffer->reopen bytes, it will close/reopen the output plugin. */
    if(the_op_config.force_reopen) {
      buffer->reopen = buffer->used;
      DEBUG(("[crossfade] open_audio: reopen=%d ms\n", B2MS(buffer->reopen)));
    }

    DEBUG(("[crossfade] open_audio: XFADE: avail=%d out=%d in=%d offset=%d skip=%d\n",
	   B2MS(avail), B2MS(out_len), B2MS(in_len), B2MS(offset), B2MS(skip)));

    if(out_len_clipped || offset_clipped)
      DEBUG(("[crossfade] open_audio: XFADE: warning: xfade params clipped!\n"));
  }

  /* calculate offset of the output plugin */
  output_offset = the_op->written_time() + B2MS(buffer->used);

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

  /* done */
  return 1;
}

void
xfade_write_audio(void *ptr, int length)
{
  gint free;
  gint ofs = 0;

  /* sanity */
  if(length <= 0) return;
  if(length & 3) {
    DEBUG(("[crossfade] write_audio: truncating %d bytes!\n", length & 3));
    length &= -4;
  }

  /* update streampos (using input format size) */
  streampos += length;

  /* convert sample format (signed-16bit-ne 44100hz stereo) */
  length = convert_flow(&convert_context, (gpointer*)&ptr, length, &in_format);

  /* convert rate */
  if(!rate_context.valid || (rate_context.in_rate != in_format.rate))
    rate_config(&rate_context, in_format.rate, out_format.rate);

  length = rate_flow(&rate_context, (gpointer*)&ptr, length);

  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  /* check if device has been closed, reopen if necessary */
  if(!output_opened) {
    DEBUG(("[crossfade] write_audio: reopening output...\n"));
    if(open_output()) {
      DEBUG(("[crossfade] write_audio: reopening failed!\n"));
      pthread_mutex_unlock(&buffer_mutex);
      return;
    }
  }

  /* reset timeout */
  gettimeofday(&last_write, NULL);

  /* calculate free space */
  free = buffer->size - buffer->used;
  if(length > free) {
    DEBUG(("[crossfade] write_audio: %d bytes truncated!\n", length-free));
    length = free;
  }

  /* kill leading gap */
  if((length > 0) && (buffer->gap > 0)) {
    gint   blen = MIN(length, buffer->gap);
    gint16   *p = ptr;
    gint  index = 0;
    
    while(index < blen) {
      gint16 left = *p++, right = *p++;
      if(ABS(left) >= buffer->gap_level) break;
      if(ABS(right) >= buffer->gap_level) break;
      index += 4;
    }
    
    buffer->gap -= index;
    length -= index;
    ptr += index;
    
    if((index < blen) || (buffer->gap == 0)) {
      gint killed = buffer->gap_len - buffer->gap;
      buffer->gap = 0;

      DEBUG(("[crossfade] write_audio: leading gap size: %d/%d ms\n",
	     B2MS(killed), B2MS(buffer->gap_len)));
      
      /* fix streampos */
      streampos -= (gint64)killed * in_format.bps / out_format.bps;
    }
  }

  /* update preload. the buffer thread will not write any
   * data to the device before preload is decreased below 1. */
  if((length > 0) && (buffer->preload > 0))
    buffer->preload -= length;

  /* fade in -- FIXME: is modifying the input/effect buffer safe? */
  if((length > 0) && (buffer->fade > 0)) {
    gint16 *p = ptr;
    gint blen = MIN(length, buffer->fade);
    gint    n;

    for(n = blen/4; n > 0; n--) {
      gfloat factor = 1.0f - (((gfloat)buffer->fade / buffer->fade_len)
			      * buffer->fade_scale);
      *p++ = (gfloat)*p * factor;
      *p++ = (gfloat)*p * factor;
      buffer->fade -= 4;
    }
  }
  
  /* mix */
  while((length > 0) && (buffer->mix > 0)) {
    gint wr_index = (buffer->rd_index + buffer->used) % buffer->size;
    gint     blen = buffer->size - wr_index;
    gint16    *p1 = buffer->data + wr_index;
    gint16    *p2 = ptr + ofs;
    gint        n;

    if(blen > length) blen = length;
    if(blen > buffer->mix) blen = buffer->mix;

    for(n = blen/2; n > 0; n--) {
      gint out = (gint)*p1 + *p2++;  /* add */
      if(out > 32767)                /* clamp */
	*p1++ = 32767;
      else if(out < -32768)
	*p1++ = -32768;
      else
	*p1++ = out;
    }
    
    buffer->used += blen;
    buffer->mix -= blen;
    length -= blen;
    ofs += blen;
  }

  /* normal write */
  while(length > 0) {
    gint wr_index = (buffer->rd_index + buffer->used) % buffer->size;
    gint     blen = buffer->size - wr_index;

    if(blen > length) blen = length;

    memcpy(buffer->data + wr_index, ptr + ofs, blen);
    
    buffer->used += blen;
    length -= blen;
    ofs += blen;
  }

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);
}

/* helper function: wait for output plugin to finish playback */
static void sync_output()
{
  glong dt, total;
  gint  opt, opt_last;
  struct timeval tv, tv_start, tv_last_change;

  if(!the_op->buffer_playing || !the_op->buffer_playing()) return;

  DEBUG(("[crossfade] sync_output: waiting for plugin...\n"));

  dt = 0;
  opt_last = 0;
  gettimeofday(&tv_start, NULL);
  gettimeofday(&tv_last_change, NULL);

  while((dt < SYNC_OUTPUT_TIMEOUT)
	&& !stopped
	&& output_opened
	&& the_op && the_op->buffer_playing()) {

    /* use output_time() to check if the output plugin is still active */
    if(the_op->output_time) {
      opt = the_op->output_time();
      if(opt != opt_last) {
	/* output_time has changed */
	opt_last = opt;
	gettimeofday(&tv_last_change, NULL);
      }
      else {
	/* calculate time since last change of the_op->output_time() */
	gettimeofday(&tv, NULL);
	dt = (tv.tv_sec - tv_last_change.tv_sec)  * 1000
	  + (tv.tv_usec - tv_last_change.tv_usec) / 1000;
      }
    }
    
    /* yield */
    pthread_mutex_unlock(&buffer_mutex);
    xmms_usleep(10000);
    pthread_mutex_lock(&buffer_mutex);
  }

  /* calculate total time we spent in here */
  gettimeofday(&tv, NULL);
  total = (tv.tv_sec - tv_start.tv_sec)  * 1000
    +    (tv.tv_usec - tv_start.tv_usec) / 1000;
  
  /* print some debug info */
  if(stopped) {
    DEBUG(("[crossfade] sync_output: ... stopped\n"));
  }
  else if(dt >= SYNC_OUTPUT_TIMEOUT) {
    DEBUG(("[crossfade] sync_output: ... TIMEOUT! (%ld ms)\n", total));
  }
  else {
    DEBUG(("[crossfade] sync_output: ... done (%ld ms)\n", total));
  }
}
	  
void *
buffer_thread_f(void *arg)
{
  gpointer data;
  gint     sync;
  gint     op_free;
  gint     length_bak, length, blen;
  glong    timeout, dt;

  struct timeval tv;
  struct timeval mark;

  DEBUG(("[crossfade] buffer_thread_f: thread started\n"));
  
  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  while(!stopped) {
    /* yield */
    pthread_mutex_unlock(&buffer_mutex);
    xmms_usleep(10000);
    pthread_mutex_lock(&buffer_mutex);

    /* get free space in device output buffer
     * NOTE: disk_writer always returns <big int> here */
    op_free = the_op->buffer_free();

    /* continue waiting if there is no room in the device buffer */
    if(op_free == 0) continue;

    /* calculate time since last write to the buffer (ignore overflows) */
    gettimeofday(&tv, NULL);
    timeout = (tv.tv_sec - last_write.tv_sec)  * 1000
      +      (tv.tv_usec - last_write.tv_usec) / 1000;
    
    /* is there any data in the buffer? */
    if(!paused && (buffer->used > 0) && (buffer->preload <= 0)) {
      /* write as much data as a) is available and b) the device can take */
      length = buffer->used;
      if(length > op_free) length = op_free;

      /* HACK: throttle output (used with fast output plugins) */
      if(the_op_config.throttle_enable && !realtime
	 && (opened || (timeout < config->songchange_timeout))) {
	sync = buffer->sync_size - (buffer->size - buffer->used);
	if(sync < 0)
	  length = 0;
	else if(sync < length)
	  length = sync;
      }

      /* clip length to reopen countdown (if applicable) */
      if((buffer->reopen >= 4) && (length > buffer->reopen))
	length = buffer->reopen;

      /* make sure we always operate on stereo sample boundary */
      length &= -4;

      /* HACK: don't stay in here too long when in realtime mode (see below) */
      if(realtime) gettimeofday(&mark, NULL);

      /* write length bytes to the device */
      length_bak = length;
      while(length > 0) {
	data = buffer->data + buffer->rd_index;
	blen = buffer->size - buffer->rd_index;
	if(blen > length) blen = length;

	/* HACK: the original OSS plugin hangs when writing large
         * blocks (greater than device buffer size) in realtime mode */
	if(the_op_config.max_write_enable
	   && (blen > the_op_config.max_write_len))
	  blen = the_op_config.max_write_len;

	/* finally, write data */
	the_op->write_audio(data, blen);

	buffer->rd_index = (buffer->rd_index + blen) % buffer->size;
	buffer->used -= blen;
	length -= blen;
	
	/* HACK: don't stay in here too long (force yielding every 10 ms) */
	if(realtime) {
	  gettimeofday(&tv, NULL);
	  dt = (tv.tv_sec - mark.tv_sec)  * 1000 
	    + (tv.tv_usec - mark.tv_usec) / 1000;
	  if(dt >= 10) break;
	}
      }

      /* calculate how many bytes actually have been written */
      length = length_bak - length;

      /* update local write accumulator */
      buffer->written += length;

      /* check reopen countdown, close/reopen output plugin when --> <=0 */
      if(buffer->reopen >= 0) {
	buffer->reopen -= length;
	if(buffer->reopen <= 0) {
 	  DEBUG(("[crossfade] buffer_thread_f: closing/reopening device\n"));
	  sync_output();

	  if(the_op->close_audio) the_op->close_audio();
	  if(!the_op->open_audio(out_format.fmt, out_format.rate, out_format.nch)) {
	    DEBUG(("[crossfade] buffer_thread_f: reopening output plugin failed!\n"));
	    g_free(buffer->data);
	    output_opened = FALSE;
	    pthread_mutex_unlock(&buffer_mutex);
	    pthread_exit(0);
	  }

	  buffer->written = 0;
	  output_offset = 0;

	  /* make sure reopen is not 0 */
	  buffer->reopen = -1;
	}
      }
    }

    /* check for special conditions */
    if(!opened) {
      if(playing) {
	if(timeout >= config->songchange_timeout) {
	  DEBUG(("[crossfade] buffer_thread_f: timeout, manual stop\n"));
	  break;
	}
      }
      else {
	if((buffer->used == 0) && (timeout >= config->songchange_timeout)) {
	  DEBUG(("[crossfade] buffer_thread_f: timeout, end of playlist\n"));
	  sync_output();
	  break;
	}
      }
    }
  }
  
  /* cleanup */
  if(output_opened) {
    pthread_mutex_unlock(&buffer_mutex);
    xfade_stop_monitor();
    pthread_mutex_lock(&buffer_mutex);

    if(the_op->close_audio) the_op->close_audio();
    g_free(buffer->data);
    output_opened = FALSE;
    DEBUG(("[crossfade] buffer_thread_f: output closed\n"));
  }
  else
    DEBUG(("[crossfade] buffer_thread_f: output already closed!\n"));

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

#if 0
  /* FIXME: this one seems to segfault somtimes when quitting xmms */
  /* final update of monitor window (clear display) */
  GDK_THREADS_ENTER();
  xfade_update_monitor(NULL);
  GDK_THREADS_LEAVE();
#endif

  DEBUG(("[crossfade] buffer_thread_f: thread finished\n"));
  pthread_exit(0);
}

void
xfade_close_audio()
{
  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  if(playing) {
    DEBUG(("[crossfade] close: stop\n"));

    /* immediatelly close output when paused */
    if(paused) {
      paused  = FALSE;
      stopped = TRUE;
    }

    fade_config = &config->fc[FADE_CONFIG_MANUAL];
  }
  else {
    DEBUG(("[crossfade] close: songchange/eop\n"));

    /* kill trailing gap (does not use buffer->gap_*) */
    if(output_opened && xfade_cfg_gap_trail_enable(config)) {
      gint gap_len   = MS2B(xfade_cfg_gap_trail_len(config)) & -4;
      gint gap_level = xfade_cfg_gap_trail_level(config);
      
      gint length = MIN(gap_len, buffer->used);
      gint killed = 0;

      while(length > 0) {
	gint wr_xedni = (buffer->rd_index + buffer->used-1) % buffer->size+1;
	gint     blen = MIN(length, wr_xedni);
	gint16     *p = buffer->data + wr_xedni;
	gint    index = 0;
	
	while(index < blen) {
	  gint16 right = *--p, left = *--p;
	  if(ABS(left) >= gap_level) break;
	  if(ABS(right) >= gap_level) break;
	  index += 4;
	}
	
	buffer->used -= index;
	killed += index;
	
	if(index < blen) break;
	length -= blen;
      }
      
      DEBUG(("[crossfade] close: trailing gap size: %d/%d ms\n",
	     B2MS(killed), B2MS(gap_len)));
    }

    fade_config = &config->fc[FADE_CONFIG_XFADE];
  }

  /* XMMS has closed the plugin */
  opened = FALSE;

  /* update last_close */
  gettimeofday(&last_close, NULL);

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);
}

void
xfade_flush(gint time)
{
  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  /* update streampos with new stream position (input format size) */
  streampos = (gint64)time * in_format.bps / 1000;

  /* flush output */
  the_op->flush(time);
  output_offset = 0;

  /* reset buffer */
  buffer->rd_index  = 0;
  buffer->used      = 0;
  buffer->preload   = buffer->preload_size;  
  buffer->mix       = 0;
  buffer->gap       = 0;
  buffer->fade      = 0;
  buffer->written   = 0;
  buffer->reopen    = -1;

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

  DEBUG(("[crossfade] flush: time=%d\n", time));
}

void
xfade_pause(short p)
{
  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  paused = p;
  the_op->pause(p);

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

  DEBUG(("[crossfade] pause: paused=%d\n", paused));
}

gint
xfade_buffer_free()
{
  gint size, free;

  /* sanity check */
  if(!output_opened) {
    DEBUG(("[crossfade] buffer_free: warning: output closed!\n"));
    return buffer->sync_size;
  }

  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  size = buffer->size;

  /* When running in realtime mode, we need to take special care here:
   * While XMMS is writing data to the output plugin, it will not yield
   * ANY processing time to the buffer thread. It will only stop when
   * xfade_buffer_free() no longer reports free buffer space. */
  if(realtime) {
    gint64 wanted = buffer->written + buffer->preload_size;

#if 1
    /* Fix for XMMS misbehaviour (possibly a bug): If the free space as
     * returned by xfade_buffer_free() is below a certain minimum block size
     * (tests showed 2304 bytes), XMMS will not send more data until there
     * is enough room for one of those blocks.
     *
     * This breaks preloading in realtime mode. To make sure that the pre-
     * load buffer gets filled we request additional sync_size bytes. */
    wanted += buffer->sync_size;
#endif

    if(wanted <= size) size = wanted;
  }

  free = size - buffer->used;
  if(free < 0) free = 0;
  
  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

  /* Convert to input format size. For input rates > output rate this will
   * return less free space than actually is available, but we don't care. */
  free /= (out_format.rate / (in_format.rate+1)) + 1;
  if(in_format.is_8bit)  free /= 2;
  if(in_format.nch == 1) free /= 2;

  /* DEBUG(("[crossfade] free: %d\n", free)); */
  return free;
}

gint
xfade_buffer_playing()
{
  /*  always return FALSE here (if not in pause) so XMMS immediatelly
   *  starts playback of the next song */

  /*
   *  NOTE: this causes trouble when playing HTTP audio streams.
   *
   *  mpg123.lib will start prebuffering (and thus stalling output) when both
   *  1) it's internal buffer is emptied (does happen all the time)
   *  2) the output plugin's buffer_playing() (this function) returns FALSE
   */

  if(paused)
    playing = TRUE;
  else
    playing = is_http && (buffer->used > 0) && the_op->buffer_playing();

  /* DEBUG(("[crossfade] playing: %d\n", playing)); */
  return playing;
}

gint
xfade_written_time()
{
  if(!output_opened) return 0;
  return streampos * 1000 / in_format.bps;
}

gint
xfade_output_time()
{
  gint time;

  /* sanity check (note: this one _does_ happen all the time) */
  if(!output_opened) return 0;

  /* lock buffer */
  pthread_mutex_lock(&buffer_mutex);

  time = the_op->output_time() - output_offset;
  if(time < 0) time = 0;

  /* unlock buffer */
  pthread_mutex_unlock(&buffer_mutex);

  return time;
}
