/* 
 *  XMMS Crossfade Plugin
 *  Copyright (C) 2000  Peter Eisenlohr <p.eisenlohr@gmx.net>
 *
 *  based on the original OSS Output Plugin
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef _CROSSFADE_H_
#define _CROSSFADE_H_

#include "config.h"

#include <pthread.h>
#include <semaphore.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include "xmms/plugin.h"
#include "xmms/xmmsctrl.h"
#include "xmms/configfile.h"
#include "xmms/util.h"

#include "debug.h"


#define OUTPUT_RATE 44100
#define OUTPUT_BPS (OUTPUT_RATE * 4)
#define OUTPUT_NCH  2

#define SYNC_OUTPUT_TIMEOUT 3000

#define OUTPUT_METHOD_BUILTIN_OSS  0
#define OUTPUT_METHOD_PLUGIN       1
#define OUTPUT_METHOD_BUILTIN_NULL 2

#define FADE_CONFIG_XFADE  0
#define FADE_CONFIG_MANUAL 1
#define MAX_FADE_CONFIGS   2

#define FADE_TYPE_REOPEN      0
#define FADE_TYPE_FLUSH       1
#define FADE_TYPE_NONE        2
#define FADE_TYPE_PAUSE       3
#define FADE_TYPE_SIMPLE_XF   4
#define FADE_TYPE_ADVANCED_XF 5
#define MAX_FADE_TYPES        6

#define FC_OFFSET_NONE     0
#define FC_OFFSET_LOCK_IN  1
#define FC_OFFSET_LOCK_OUT 2
#define FC_OFFSET_CUSTOM   3

#define DEFAULT_OSS_ALT_AUDIO_DEVICE "/dev/dsp"
#define DEFAULT_OSS_ALT_MIXER_DEVICE "/dev/mixer"
#define DEFAULT_OP_CONFIG_STRING     "libOSS.so=0,1,2304,0; libdisk_writer.so=1,0,2304,1"
#define DEFAULT_OP_NAME              "libOSS.so"

#define DEFAULT_OP_CONFIG			\
{ FALSE, FALSE, 2304, FALSE }

#define DEFAULT_FADE_CONFIG			\
{ FADE_TYPE_NONE,				\
  2000,						\
  4000,						\
  FALSE, 1500, 0,				\
  FC_OFFSET_NONE,				\
  FC_OFFSET_NONE, -500,				\
  TRUE,						\
  FALSE, 1500, 0,				\
  FALSE						\
}

#define CONFIG_DEFAULT				\
{ 0,            /* output_method */		\
  0,            /* oss_audio_device */		\
  FALSE,        /* oss_use_alt_audio_device */	\
  NULL,         /* oss_alt_audio_device */	\
  0,            /* oss_mixer_device */		\
  FALSE,        /* oss_use_alt_mixer_device */	\
  NULL,         /* oss_alt_mixer_device */	\
  FALSE,        /* oss_mixer_use_master */	\
  0,            /* oss_buffer_size_ms */	\
  500,          /* oss_preload_size_ms */	\
  NULL,         /* op_config_string */		\
  NULL,         /* op_name */			\
  10000,        /* mix_size_ms */		\
  TRUE,         /* mix_size_auto */		\
  {             /* fc[MAX_FADE_CONFIGS] */	\
    { FADE_TYPE_ADVANCED_XF,			\
      2000,					\
      3000,					\
      TRUE, 6000, 0,				\
      FC_OFFSET_CUSTOM,				\
      FC_OFFSET_CUSTOM, -8000,			\
      TRUE,					\
      FALSE, 4000, 33,				\
      FALSE					\
    },						\
    { FADE_TYPE_NONE,				\
      2000,					\
      1000,					\
      TRUE, 6000, 0,				\
      FC_OFFSET_LOCK_OUT,			\
      FC_OFFSET_LOCK_OUT, -8000,		\
      TRUE,					\
      FALSE, 4000, 33,				\
      TRUE					\
    }						\
  },						\
  TRUE,         /* gap_lead_enable */		\
  500,          /* gap_lead_len_ms */		\
  512,          /* gap_lead_level */		\
  TRUE,         /* gap_trail_enable */		\
  500,          /* gap_trail_len_ms */		\
  512,          /* gap_trail_level */		\
  TRUE,         /* gap_trail_locked */		\
  FALSE,        /* enable_debug */		\
  FALSE,        /* enable_monitor */		\
  TRUE,         /* enable_mixer */		\
  FALSE,        /* mixer_reverse */		\
  500,          /* songchange_timeout */	\
  250,          /* sync_size_ms */		\
  0,            /* preload_size_ms */		\
  TRUE          /* enable_http_workaround */	\
}


#define DEBUG(x)  if(config->enable_debug) debug x
#define PERROR(x) if(config->enable_debug) perror(x)

#define WRAP(x,n) (((x)<0)?((n)-(x))%(n):((x)%(n)))
#define B2MS(x)   ((gint)((gint64)(x)*1000/OUTPUT_BPS))
#define MS2B(x)   ((gint)((gint64)(x)*OUTPUT_BPS/1000))


/* get plugin info (imported by XMMS) */
OutputPlugin *get_oplugin_info(void);

/* xmms internal prototypes */
gint   ctrlsocket_get_session_id(void);
GList *get_output_list(void);


typedef struct
{
  gpointer data;      /* buffer */
  gint     size;      /* total buffer length */

  gint used;          /* length */
  gint rd_index;      /* offset */

  gint64 written;     /* bytes written since last open/flush */
  
  gint mix_size;      /* mixing buffer length */
  gint sync_size;     /* additional buffer space for mix timing */
  gint preload_size;  /* preload buffer length */

  gint mix;           /* > 0: mixing new data into buffer */
  gint preload;       /* > 0: preloading */

  gint gap;           /* > 0: removing (leading) gap */
  gint gap_len;       /* max. len of gap, 0=disabled */
  gint gap_level;     /* max. sample value+1 to be considered "silent" */

  gint   fade;        /* > 0: fade in new data (xfade only) */
  gint   fade_len;    /* length of fade */
  gfloat fade_scale;  /* 1.0f - in_level */

  gint   reopen;      /* > 0: countdown to reopen device (disk_writer hack) */
}
buffer_t;

typedef struct
{
  gint     type;

  gint     pause_len_ms;     // PAUSE

  gint     simple_len_ms;    // SIMPLE_XF

  gboolean out_enable;       // ADVANCED_XF
  gint     out_len_ms;       // ADVANCED_XF
  gint     out_volume;       // ADVANCED_XF

  gint     ofs_type;         // ADVANCED_XF
  gint     ofs_type_wanted;  // ADVANCED_XF
  gint     ofs_custom_ms;    // ADVANCED_XF

  gboolean in_locked;        // ADVANCED_XF
  gboolean in_enable;        // ADVANCED_XF
  gint     in_len_ms;        // ADVANCED_XF
  gint     in_volume;        // ADVANCED_XF

  gboolean flush;            // TRUE for manual, FALSE for xfade config
}
fade_config_t;

typedef struct
{
  gboolean  throttle_enable;
  gboolean  max_write_enable;
  gint      max_write_len;
  gboolean  force_reopen;
}
plugin_config_t;

typedef struct
{
  /* output method */
  gint output_method;

  /* output method: builtin OSS */
  gint     oss_audio_device;
  gboolean oss_use_alt_audio_device;
  gchar   *oss_alt_audio_device;

  gint     oss_mixer_device;
  gboolean oss_use_alt_mixer_device;
  gchar   *oss_alt_mixer_device;
  gboolean oss_mixer_use_master;

  gint     oss_buffer_size_ms;
  gint     oss_preload_size_ms;

  /* output method: plugin */
  gchar          *op_config_string;  /* stores configs for all plugins */
  gchar          *op_name;           /* name of the current plugin */

  /* crossfader */
  gint          mix_size_ms;
  gboolean      mix_size_auto;
  fade_config_t fc[MAX_FADE_CONFIGS];

  /* gap killer */
  gboolean gap_lead_enable;
  gint     gap_lead_len_ms;
  gint     gap_lead_level;
  gboolean gap_trail_enable;
  gint     gap_trail_len_ms;
  gint     gap_trail_level;
  gboolean gap_trail_locked;

  /* misc */
  gboolean enable_debug;
  gboolean enable_monitor;

  gboolean enable_mixer;
  gboolean mixer_reverse;

  gint     songchange_timeout;

  /* additional stuff which is not configureable by the user */
  gint     sync_size_ms;
  gint     preload_size_ms;
  gboolean enable_http_workaround;

  /* additional stuff which is not saved to the config file */
  gint     page;
  gint     oss_page;
  gint     xf_index;
}
config_t;


/* some global vars... we should really get rid of those somehow */
extern config_t *config;
extern config_t  config_default;
extern buffer_t *buffer;

extern pthread_mutex_t buffer_mutex;

extern gboolean opened;         /* XMMS-crossfade is opened */
extern gboolean output_opened;  /* the output plugin is opened */

extern OutputPlugin *the_op;

#endif  /* _CROSSFADE_H_ */
